import pygame


class Keys:
    """
    random object structure to handle keyboard keys
    """
    def __init__(self) -> None:
        pass

    @property
    def K_BACKSPACE(self):
        """
        backspace
        """
        return pygame.K_BACKSPACE

    @property
    def K_TAB(self):
        """
        tab
        """
        return pygame.K_TAB

    @property
    def K_CLEAR(self):
        """
        clear
        """
        return pygame.K_CLEAR

    @property
    def K_RETURN(self):
        """
        return
        """
        return pygame.K_RETURN

    @property
    def K_PAUSE(self):
        """
        pause
        """
        return pygame.K_PAUSE

    @property
    def K_ESCAPE(self):
        """
        escape
        """
        return pygame.K_ESCAPE

    @property
    def K_SPACE(self):
        """
        space
        """
        return pygame.K_SPACE

    @property
    def K_EXCLAIM(self):
        """
        exclaim
        """
        return pygame.K_EXCLAIM

    @property
    def K_QUOTEDBL(self):
        """
        quotedbl
        """
        return pygame.K_QUOTEDBL

    @property
    def K_HASH(self):
        """
        hash
        """
        return pygame.K_HASH

    @property
    def K_DOLLAR(self):
        """
        dollar
        """
        return pygame.K_DOLLAR

    @property
    def K_AMPERSAND(self):
        """
        ampersand
        """
        return pygame.K_AMPERSAND

    @property
    def K_QUOTE(self):
        """
        quote
        """
        return pygame.K_QUOTE

    @property
    def K_LEFTPAREN(self):
        """
        left parenthesis
        """
        return pygame.K_LEFTPAREN

    @property
    def K_RIGHTPAREN(self):
        """
        right parenthesis
        """
        return pygame.K_RIGHTPAREN

    @property
    def K_ASTERISK(self):
        """
        asterisk
        """
        return pygame.K_ASTERISK

    @property
    def K_PLUS(self):
        """
        plus sign
        """
        return pygame.K_PLUS

    @property
    def K_COMMA(self):
        """
        comma
        """
        return pygame.K_COMMA

    @property
    def K_MINUS(self):
        """
        minus sign
        """
        return pygame.K_MINUS

    @property
    def K_PERIOD(self):
        """
        period
        """
        return pygame.K_PERIOD

    @property
    def K_SLASH(self):
        """
        forward slash
        """
        return pygame.K_SLASH

    @property
    def K_0(self):
        """
        0
        """
        return pygame.K_0

    @property
    def K_1(self):
        """
        1
        """
        return pygame.K_1

    @property
    def K_2(self):
        """
        2
        """
        return pygame.K_2

    @property
    def K_3(self):
        """
        3
        """
        return pygame.K_3

    @property
    def K_4(self):
        """
        4
        """
        return pygame.K_4

    @property
    def K_5(self):
        """
        5
        """
        return pygame.K_5

    @property
    def K_6(self):
        """
        6
        """
        return pygame.K_6

    @property
    def K_7(self):
        """
        7
        """
        return pygame.K_7

    @property
    def K_8(self):
        """
        8
        """
        return pygame.K_8

    @property
    def K_9(self):
        """
        9
        """
        return pygame.K_9

    @property
    def K_COLON(self):
        """
        colon
        """
        return pygame.K_COLON

    @property
    def K_SEMICOLON(self):
        """
        semicolon
        """
        return pygame.K_SEMICOLON

    @property
    def K_LESS(self):
        """
        less-than sign
        """
        return pygame.K_LESS

    @property
    def K_EQUALS(self):
        """
        equals sign
        """
        return pygame.K_EQUALS

    @property
    def K_GREATER(self):
        """
        greater-than sign
        """
        return pygame.K_GREATER

    @property
    def K_QUESTION(self):
        """
        question mark
        """
        return pygame.K_QUESTION

    @property
    def K_AT(self):
        """
        at
        """
        return pygame.K_AT

    @property
    def K_LEFTBRACKET(self):
        """
        left bracket
        """
        return pygame.K_LEFTBRACKET

    @property
    def K_BACKSLASH(self):
        """
        backslash
        """
        return pygame.K_BACKSLASH

    @property
    def K_RIGHTBRACKET(self):
        """
        right bracket
        """
        return pygame.K_RIGHTBRACKET

    @property
    def K_CARET(self):
        """
        caret
        """
        return pygame.K_CARET

    @property
    def K_UNDERSCORE(self):
        """
        underscore
        """
        return pygame.K_UNDERSCORE

    @property
    def K_BACKQUOTE(self):
        """
        grave
        """
        return pygame.K_BACKQUOTE

    @property
    def K_a(self):
        """
        a
        """
        return pygame.K_a

    @property
    def K_b(self):
        """
        b
        """
        return pygame.K_b

    @property
    def K_c(self):
        """
        c
        """
        return pygame.K_c

    @property
    def K_d(self):
        """
        d
        """
        return pygame.K_d

    @property
    def K_e(self):
        """
        e
        """
        return pygame.K_e

    @property
    def K_f(self):
        """
        f
        """
        return pygame.K_f

    @property
    def K_g(self):
        """
        g
        """
        return pygame.K_g

    @property
    def K_h(self):
        """
        h
        """
        return pygame.K_h

    @property
    def K_i(self):
        """
        i
        """
        return pygame.K_i

    @property
    def K_j(self):
        """
        j
        """
        return pygame.K_j

    @property
    def K_k(self):
        """
        k
        """
        return pygame.K_k

    @property
    def K_l(self):
        """
        l
        """
        return pygame.K_l

    @property
    def K_m(self):
        """
        m
        """
        return pygame.K_m

    @property
    def K_n(self):
        """
        n
        """
        return pygame.K_n

    @property
    def K_o(self):
        """
        o
        """
        return pygame.K_o

    @property
    def K_p(self):
        """
        p
        """
        return pygame.K_p

    @property
    def K_q(self):
        """
        q
        """
        return pygame.K_q

    @property
    def K_r(self):
        """
        r
        """
        return pygame.K_r

    @property
    def K_s(self):
        """
        s
        """
        return pygame.K_s

    @property
    def K_t(self):
        """
        t
        """
        return pygame.K_t

    @property
    def K_u(self):
        """
        u
        """
        return pygame.K_u

    @property
    def K_v(self):
        """
        v
        """
        return pygame.K_v

    @property
    def K_w(self):
        """
        w
        """
        return pygame.K_w

    @property
    def K_x(self):
        """
        x
        """
        return pygame.K_x

    @property
    def K_y(self):
        """
        y
        """
        return pygame.K_y

    @property
    def K_z(self):
        """
        z
        """
        return pygame.K_z

    @property
    def K_DELETE(self):
        """
        delete
        """
        return pygame.K_DELETE

    @property
    def K_KP0(self):
        """
        keypad 0
        """
        return pygame.K_KP0

    @property
    def K_KP1(self):
        """
        keypad 1
        """
        return pygame.K_KP1

    @property
    def K_KP2(self):
        """
        keypad 2
        """
        return pygame.K_KP2

    @property
    def K_KP3(self):
        """
        keypad 3
        """
        return pygame.K_KP3

    @property
    def K_KP4(self):
        """
        keypad 4
        """
        return pygame.K_KP4

    @property
    def K_KP5(self):
        """
        keypad 5
        """
        return pygame.K_KP5

    @property
    def K_KP6(self):
        """
        keypad 6
        """
        return pygame.K_KP6

    @property
    def K_KP7(self):
        """
        keypad 7
        """
        return pygame.K_KP7

    @property
    def K_KP8(self):
        """
        keypad 8
        """
        return pygame.K_KP8

    @property
    def K_KP9(self):
        """
        keypad 9
        """
        return pygame.K_KP9

    @property
    def K_KP_PERIOD(self):
        """
        keypad period
        """
        return pygame.K_KP_PERIOD

    @property
    def K_KP_DIVIDE(self):
        """
        keypad divide
        """
        return pygame.K_KP_DIVIDE

    @property
    def K_KP_MULTIPLY(self):
        """
        keypad multiply
        """
        return pygame.K_KP_MULTIPLY

    @property
    def K_KP_MINUS(self):
        """
        keypad minus
        """
        return pygame.K_KP_MINUS

    @property
    def K_KP_PLUS(self):
        """
        keypad plus
        """
        return pygame.K_KP_PLUS

    @property
    def K_KP_ENTER(self):
        """
        keypad enter
        """
        return pygame.K_KP_ENTER

    @property
    def K_KP_EQUALS(self):
        """
        keypad equals
        """
        return pygame.K_KP_EQUALS

    @property
    def K_UP(self):
        """
        up arrow
        """
        return pygame.K_UP

    @property
    def K_DOWN(self):
        """
        down arrow
        """
        return pygame.K_DOWN

    @property
    def K_RIGHT(self):
        """
        right arrow
        """
        return pygame.K_RIGHT

    @property
    def K_LEFT(self):
        """
        left arrow
        """
        return pygame.K_LEFT

    @property
    def K_INSERT(self):
        """
        insert
        """
        return pygame.K_INSERT

    @property
    def K_HOME(self):
        """
        home
        """
        return pygame.K_HOME

    @property
    def K_END(self):
        """
        end
        """
        return pygame.K_END

    @property
    def K_PAGEUP(self):
        """
        page up
        """
        return pygame.K_PAGEUP

    @property
    def K_PAGEDOWN(self):
        """
        page down
        """
        return pygame.K_PAGEDOWN

    @property
    def K_F1(self):
        """
        F1
        """
        return pygame.K_F1

    @property
    def K_F2(self):
        """
        F2
        """
        return pygame.K_F2

    @property
    def K_F3(self):
        """
        F3
        """
        return pygame.K_F3

    @property
    def K_F4(self):
        """
        F4
        """
        return pygame.K_F4

    @property
    def K_F5(self):
        """
        F5
        """
        return pygame.K_F5

    @property
    def K_F6(self):
        """
        F6
        """
        return pygame.K_F6

    @property
    def K_F7(self):
        """
        F7
        """
        return pygame.K_F7

    @property
    def K_F8(self):
        """
        F8
        """
        return pygame.K_F8

    @property
    def K_F9(self):
        """
        F9
        """
        return pygame.K_F9

    @property
    def K_F10(self):
        """
        F10
        """
        return pygame.K_F10

    @property
    def K_F11(self):
        """
        F11
        """
        return pygame.K_F11

    @property
    def K_F12(self):
        """
        F12
        """
        return pygame.K_F12

    @property
    def K_F13(self):
        """
        F13
        """
        return pygame.K_F13

    @property
    def K_F14(self):
        """
        F14
        """
        return pygame.K_F14

    @property
    def K_F15(self):
        """
        F15
        """
        return pygame.K_F15

    @property
    def K_NUMLOCK(self):
        """
        numlock
        """
        return pygame.K_NUMLOCK

    @property
    def K_CAPSLOCK(self):
        """
        capslock
        """
        return pygame.K_CAPSLOCK

    @property
    def K_SCROLLOCK(self):
        """
        scrollock
        """
        return pygame.K_SCROLLOCK

    @property
    def K_RSHIFT(self):
        """
        right shift
        """
        return pygame.K_RSHIFT

    @property
    def K_LSHIFT(self):
        """
        left shift
        """
        return pygame.K_LSHIFT

    @property
    def K_RCTRL(self):
        """
        right control
        """
        return pygame.K_RCTRL

    @property
    def K_LCTRL(self):
        """
        left control
        """
        return pygame.K_LCTRL

    @property
    def K_RALT(self):
        """
        right alt
        """
        return pygame.K_RALT

    @property
    def K_LALT(self):
        """
        left alt
        """
        return pygame.K_LALT

    @property
    def K_RMETA(self):
        """
        right meta
        """
        return pygame.K_RMETA

    @property
    def K_LMETA(self):
        """
        left meta
        """
        return pygame.K_LMETA

    @property
    def K_LSUPER(self):
        """
        left Windows key
        """
        return pygame.K_LSUPER

    @property
    def K_RSUPER(self):
        """
        right Windows key
        """
        return pygame.K_RSUPER

    @property
    def K_MODE(self):
        """
        mode shift
        """
        return pygame.K_MODE

    @property
    def K_HELP(self):
        """
        help
        """
        return pygame.K_HELP

    @property
    def K_PRINT(self):
        """
        print screen
        """
        return pygame.K_PRINT

    @property
    def K_SYSREQ(self):
        """
        sysrq
        """
        return pygame.K_SYSREQ

    @property
    def K_BREAK(self):
        """
        break
        """
        return pygame.K_BREAK

    @property
    def K_MENU(self):
        """
        menu
        """
        return pygame.K_MENU

    @property
    def K_POWER(self):
        """
        power
        """
        return pygame.K_POWER

    @property
    def K_EURO(self):
        """
        Euro
        """
        return pygame.K_EURO


all_keys = [
    "K_BACKSPACE", "K_TAB", "K_CLEAR", "K_RETURN", "K_PAUSE", "K_ESCAPE",
    "K_SPACE", "K_EXCLAIM", "K_QUOTEDBL", "K_HASH", "K_DOLLAR", "K_AMPERSAND",
    "K_QUOTE", "K_LEFTPAREN", "K_RIGHTPAREN", "K_ASTERISK", "K_PLUS",
    "K_COMMA", "K_MINUS", "K_PERIOD", "K_SLASH", "K_0", "K_1", "K_2", "K_3",
    "K_4", "K_5", "K_6", "K_7", "K_8", "K_9", "K_COLON", "K_SEMICOLON",
    "K_LESS", "K_EQUALS", "K_GREATER", "K_QUESTION", "K_AT", "K_LEFTBRACKET",
    "K_BACKSLASH", "K_RIGHTBRACKET", "K_CARET", "K_UNDERSCORE", "K_BACKQUOTE",
    "K_a", "K_b", "K_c", "K_d", "K_e", "K_f", "K_g", "K_h", "K_i", "K_j",
    "K_k", "K_l", "K_m", "K_n", "K_o", "K_p", "K_q", "K_r", "K_s", "K_t",
    "K_u", "K_v", "K_w", "K_x", "K_y", "K_z", "K_DELETE", "K_KP0", "K_KP1",
    "K_KP2", "K_KP3", "K_KP4", "K_KP5", "K_KP6", "K_KP7", "K_KP8", "K_KP9",
    "K_KP_PERIOD", "K_KP_DIVIDE", "K_KP_MULTIPLY", "K_KP_MINUS", "K_KP_PLUS",
    "K_KP_ENTER", "K_KP_EQUALS", "K_UP", "K_DOWN", "K_RIGHT", "K_LEFT",
    "K_INSERT", "K_HOME", "K_END", "K_PAGEUP", "K_PAGEDOWN", "K_F1", "K_F2",
    "K_F3", "K_F4", "K_F5", "K_F6", "K_F7", "K_F8", "K_F9", "K_F10", "K_F11",
    "K_F12", "K_F13", "K_F14", "K_F15", "K_NUMLOCK", "K_CAPSLOCK",
    "K_SCROLLOCK", "K_RSHIFT", "K_LSHIFT", "K_RCTRL", "K_LCTRL", "K_RALT",
    "K_LALT", "K_RMETA", "K_LMETA", "K_LSUPER", "K_RSUPER", "K_MODE", "K_HELP",
    "K_PRINT", "K_SYSREQ", "K_BREAK", "K_MENU", "K_POWER", "K_EURO"
]

pygame_key_binding = [
    pygame.K_BACKSPACE, pygame.K_TAB, pygame.K_CLEAR, pygame.K_RETURN,
    pygame.K_PAUSE, pygame.K_ESCAPE, pygame.K_SPACE, pygame.K_EXCLAIM,
    pygame.K_QUOTEDBL, pygame.K_HASH, pygame.K_DOLLAR, pygame.K_AMPERSAND,
    pygame.K_QUOTE, pygame.K_LEFTPAREN, pygame.K_RIGHTPAREN, pygame.K_ASTERISK,
    pygame.K_PLUS, pygame.K_COMMA, pygame.K_MINUS, pygame.K_PERIOD,
    pygame.K_SLASH, pygame.K_0, pygame.K_1, pygame.K_2, pygame.K_3, pygame.K_4,
    pygame.K_5, pygame.K_6, pygame.K_7, pygame.K_8, pygame.K_9, pygame.K_COLON,
    pygame.K_SEMICOLON, pygame.K_LESS, pygame.K_EQUALS, pygame.K_GREATER,
    pygame.K_QUESTION, pygame.K_AT, pygame.K_LEFTBRACKET, pygame.K_BACKSLASH,
    pygame.K_RIGHTBRACKET, pygame.K_CARET, pygame.K_UNDERSCORE,
    pygame.K_BACKQUOTE, pygame.K_a, pygame.K_b, pygame.K_c, pygame.K_d,
    pygame.K_e, pygame.K_f, pygame.K_g, pygame.K_h, pygame.K_i, pygame.K_j,
    pygame.K_k, pygame.K_l, pygame.K_m, pygame.K_n, pygame.K_o, pygame.K_p,
    pygame.K_q, pygame.K_r, pygame.K_s, pygame.K_t, pygame.K_u, pygame.K_v,
    pygame.K_w, pygame.K_x, pygame.K_y, pygame.K_z, pygame.K_DELETE,
    pygame.K_KP0, pygame.K_KP1, pygame.K_KP2, pygame.K_KP3, pygame.K_KP4,
    pygame.K_KP5, pygame.K_KP6, pygame.K_KP7, pygame.K_KP8, pygame.K_KP9,
    pygame.K_KP_PERIOD, pygame.K_KP_DIVIDE, pygame.K_KP_MULTIPLY,
    pygame.K_KP_MINUS, pygame.K_KP_PLUS, pygame.K_KP_ENTER, pygame.K_KP_EQUALS,
    pygame.K_UP, pygame.K_DOWN, pygame.K_RIGHT, pygame.K_LEFT, pygame.K_INSERT,
    pygame.K_HOME, pygame.K_END, pygame.K_PAGEUP, pygame.K_PAGEDOWN,
    pygame.K_F1, pygame.K_F2, pygame.K_F3, pygame.K_F4, pygame.K_F5,
    pygame.K_F6, pygame.K_F7, pygame.K_F8, pygame.K_F9, pygame.K_F10,
    pygame.K_F11, pygame.K_F12, pygame.K_F13, pygame.K_F14, pygame.K_F15,
    pygame.K_NUMLOCK, pygame.K_CAPSLOCK, pygame.K_SCROLLOCK, pygame.K_RSHIFT,
    pygame.K_LSHIFT, pygame.K_RCTRL, pygame.K_LCTRL, pygame.K_RALT,
    pygame.K_LALT, pygame.K_RMETA, pygame.K_LMETA, pygame.K_LSUPER,
    pygame.K_RSUPER, pygame.K_MODE, pygame.K_HELP, pygame.K_PRINT,
    pygame.K_SYSREQ, pygame.K_BREAK, pygame.K_MENU, pygame.K_POWER,
    pygame.K_EURO
]



from typing import Any, Callable, Union
from ..data import *

__all__ = ["Menu"]

import difflib
import pygame


def _map(x: float, x0: float, x1: float, y0: float, y1: float) -> float:
    """
    linear interpolation
    """
    return (y0 * (x1-x) + y1 * (x-x0)) / (x1-x0)


def _constrain(x: float, mn: float, mx: float) -> int:
    if x < mn:
        return mn
    elif x > mx:
        return mx
    return x


class Menu:
    """
    Pygame Menu
    ===========
    created by ``Renderer``
    """
    def __init__(self,
                 renderer,
                 name: str,
                 side: str = RIGHT,
                 background: Union[bool, tuple[int, int, int], int,
                                   str] = True,
                 length: int = None,
                 color: Union[tuple[int, int, int], int,
                              str] = (155, 155, 155),
                 text_color: Union[tuple[int, int, int], int,
                                   str] = (255, 255, 255),
                 text_size: int = 15,
                 **kwargs: Callable[[], None]) -> None:
        """
        new Menu instance

        Parameters
        ----------
            renderer : Renderer
                the Renderer instance the Menu is linked to
            name : str
                name of the menu
            side : str, (optional)
                side of the window to display the menu
                LEFT | RIGHT
                defaults to RIGHT
            background : bool | tuple | int | str, (optional)
                draw a background when expanded
                defaults to True
            length : int, (optional)
                length of the menu (its height)
                by default the menu height will on auto
                defaults to None
            color : tuple | int | str, (optional)
                lines color used for drawing when menu is visible
                defaults to (155, 155, 155)
            text_color : tuple | int | str, (optional)
                text color used for display text inside the menu
                defaults to (255, 255, 255)
            text_size : int
                text size of the menu, must be between 1 and 25

        Keywords Arguments
        ------------------
            * : str\\
                name of the buttons on the menu, in order\\
                must be linked to a python function
        """
        self.has_error = False

        self._renderer = renderer
        self._name = name
        self._namew = 100

        self._side = side
        self._length = length
        self._width = 0

        self._background = None
        self._has_background = False
        if background is None:
            self._has_background = False
        if isinstance(background, bool):
            self._has_background = background
        elif isinstance(background, tuple) and len(background) == 3:
            self._background = background
            self._has_background = True
        elif isinstance(background, int):
            self._background = background, background, background
            self._has_background = True
        elif isinstance(background, str):
            try:
                self._background = COLORS[background.lower()]
                self._has_background = True
            except KeyError:
                close = difflib.get_close_matches(background.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [menu {self._name} : {background} is not a valid color name, using closest match {close} instead"
                )
                self._background = COLORS[close]
                self._has_background = True
        else:
            warn(
                f"ERROR [menu {self._name}] : wrong background parameter, menu was not created"
            )
            self.has_error = True

        if isinstance(color, tuple) and len(color) == 3:
            self._color = color
        elif isinstance(color, int):
            self._color = color, color, color
        elif isinstance(color, str):
            try:
                self._color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [menu {self._name}] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._color = COLORS[close]
        else:
            warn(
                f"ERROR [menu {self._name}] : wrong color parameter, menu was not created"
            )
            self.has_error = True

        if isinstance(text_color, tuple) and len(text_color) == 3:
            self._text_color = text_color
        elif isinstance(text_color, int):
            self._text_color = text_color, text_color, text_color
        elif isinstance(text_color, str):
            try:
                self._text_color = COLORS[text_color.lower()]
            except KeyError:
                close = difflib.get_close_matches(text_color,
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [menu {self._name}] : {text_color} is not a valid color name, using closest match {close} instead"
                )
                self._text_color = COLORS[close]
        else:
            warn(
                f"ERROR [slider {self._name}] : wrong text color parameter, menu was not created"
            )
            self.has_error = True

        self._text_size = _constrain(text_size, 1, 25)
        self._all_items = []
        self._all_actions: list[Callable[[], None]] = []
        for k, v in kwargs.items():
            self._all_items.append(k)
            self._all_actions.append(v)

        self._click = 0
        self._click_count = 1
        self._is_hidden = False

        self._is_fold = True
        self._is_playing = False
        self._tick_count = 0
        self._max_ticks = 1

        self.set_max_width()

    def click(self) -> None:
        """
        increments click
        """
        self._click += 1

    def reinit_click(self):
        """
        puts click back to 0
        """
        self._click = 0

    def check_click(self) -> bool:
        """
        check if click greater that click_count
        """
        return self._click >= self._click_count

    @property
    def is_hidden(self) -> bool:
        """
        gets current display state of the menu
        """
        return self._is_hidden

    @is_hidden.setter
    def is_hidden(self, value: bool) -> None:
        """
        forces the display state of the menu
        """
        self._is_hidden = value

    def hide(self) -> None:
        """
        hides the menu (does not display it automatically on the screen)\\
        menu actions become unaccessible\\
        opposite method is ``reveal``
        """
        if self._is_hidden:
            warn(
                f"WARNING [menu {self._name}] : menu is already hidden, nothing changed"
            )
            return
        self._is_hidden = True

    def reveal(self) -> None:
        """
        reveals the menu back (displays it on the screen)\\
        menu actions become accessible again\\
        opposite method is ``hide``
        """
        if not self._is_hidden:
            warn(
                f"WARNING [menu {self._name}] : menu is not hidden, nothing changed"
            )
            return
        self._is_hidden = False

    def move_to(self, side: str) -> None:
        """
        moves menu to the designated location

        Parameters
        ----------
            side : str
                side of the window
                LEFT | RIGHT
        """
        if side not in (LEFT, RIGHT):
            warn(
                f"ERROR [menu {self._name}] : {side} is not a valid location, nothing changed"
            )
            return
        self._side = side

    @property
    def has_background(self) -> bool:
        """
        gets current menu background drawing
        """
        return self._has_background

    @has_background.setter
    def has_background(self, background: bool) -> None:
        """
        sets current menu background drawing

        Parameters
        ----------
            background : bool
                draw a background when expanded
        """
        self._has_background = background

    @property
    def background(self) -> tuple:
        """
        gets current menu background color\\
        might be None if background color is the windows background color
        or if background is disabled
        """
        return self._background

    @background.setter
    def background(
            self,
            background: Union[tuple[int, int, int], int, str] = False) -> None:
        """
        sets current background drawing

        Parameters
        ----------
            background : None | bool | tuple | int | str
                background draw
        """
        self._background = None
        if background is None:
            self._has_background = False
        if isinstance(background, bool):
            self._has_background = background
        elif isinstance(background, tuple) and len(background) == 3:
            self._background = background
            self._has_background = True
        elif isinstance(background, int):
            self._background = background, background, background
            self._has_background = True
        elif isinstance(background, str):
            try:
                self._background = COLORS[background.lower()]
                self._has_background = True
            except KeyError:
                close = difflib.get_close_matches(background.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [menu {self._name} : {background} is not a valid color name, using closest match {close} instead"
                )
                self._background = COLORS[close]
                self._has_background = True
        else:
            warn(
                f"ERROR [menu {self._name}] : {background} is not a valid background parameter, nothing changed"
            )

    @property
    def color(self) -> tuple:
        """
        gets current menu line color
        """
        return self._color

    @color.setter
    def color(self, color: Union[tuple[int, int, int], int, str]) -> None:
        """
        sets menu line color\\
        deprecated : do not use

        Parameters
        ----------
            color : tuple | int | str
                the new color
        """
        warn(f"INFO  [menu, {self._name}] : attempting color change")
        if isinstance(color, tuple) and len(color) == 3:
            self._color = color
        elif isinstance(color, int):
            self._color = color, color, color
        elif isinstance(color, str):
            try:
                self._color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [menu {self._name}] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._color = COLORS[close]
        else:
            warn(
                f"ERROR [menu {self._name}] : {color} is not a valid color, nothing changed"
            )

    @property
    def text_color(self) -> tuple:
        """
        gets current menu text color
        """
        return self._text_color

    @text_color.setter
    def text_color(self, text_color: Union[tuple[int, int, int], int,
                                           str]) -> None:
        """
        sets menu text color\\
        deprecated : do not use

        Parameters
        ----------
            text_color : tuple | int | str
                the new color
        """
        warn(f"INFO  [menu, {self._name}] : attempting text color change")
        if isinstance(text_color, tuple) and len(text_color) == 3:
            self._text_color = text_color
        elif isinstance(text_color, int):
            self._text_color = text_color, text_color, text_color
        elif isinstance(text_color, str):
            try:
                self._text_color = COLORS[text_color.lower()]
            except KeyError:
                close = difflib.get_close_matches(text_color,
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [menu {self._name}] : {text_color} is not a valid color name, using closest match {close} instead"
                )
                self._text_color = COLORS[close]
        else:
            warn(
                f"ERROR [menu {self._name}] : {text_color} is not a valid color, nothing changed"
            )

    @property
    def side(self) -> str:
        """
        gets current menu side
        """
        return self._side

    @side.setter
    def side(self, side: str) -> None:
        """
        moves menu to the designated location\\
        alias for ``move_to``
        """
        self.move_to(side)

    @property
    def width(self) -> int:
        """
        gets maximum width of the menu depending on its text\\
        might be off if ``set_max_width`` was not called when updating menu
        """
        return self._width

    @width.setter
    def width(self, width: int) -> None:
        """
        sets maximum width of the menu\\
        does not try to get the desired width
        """
        self._width = width

    @property
    def length(self) -> int:
        """
        gets current height draw of menu\\
        might be None
        """
        return self._length

    @length.setter
    def length(self, length) -> None:
        """
        sets current height draw of menu
        
        Parameters
        ----------
            length : None | int
                new height
        """
        self._length = length

    @property
    def name(self) -> str:
        """
        gets current menu name
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """
        sets new menu name\\
        deprecated : do not use

        Parameters
        ----------
            name : str
                new name
        """
        warn(f"INFO [button {self._name}] : name changing to {name}")
        self._name = name

    @property
    def is_fold(self) -> bool:
        """
        gets current state of the menu
        """
        return self._is_fold

    @is_fold.setter
    def is_fold(self, state: bool) -> None:
        """
        sets state of the menu
        """
        self._is_fold = state

    @property
    def tick_count(self) -> int:
        """
        gets current menu tick count (frame of animation)
        """
        return self._tick_count

    @tick_count.setter
    def tick_count(self, tick: int) -> None:
        """
        sets menu tick count
        """
        self._tick_count = tick

    @property
    def max_ticks(self) -> int:
        """
        gets current menu max ticks (max number of frames to animate)
        """
        return self._max_ticks

    @max_ticks.setter
    def max_ticks(self, max: int) -> None:
        """
        sets menu max ticks
        """
        self._max_ticks = max + 1

    @property
    def is_playing(self) -> bool:
        """
        gets current state of animation (either animating or resting)
        """
        return self._is_playing

    @is_playing.setter
    def is_playing(self, state: bool) -> None:
        """
        sets state of animation
        """
        self._is_playing = state

    @property
    def text_size(self) -> int:
        """
        gets current text size
        """
        return self._text_size

    @text_size.setter
    def text_size(self, size: int) -> None:
        """
        sets menu text size

        Parameters
        ----------
            size : int
                new text size, must be between 1 and 25, inclusive
        """
        if size > 25:
            warn(
                f"ERROR [menu {self._name}] : text size is too big, max is 25 due to menu layout, nothing changed"
            )
            return
        if size < 1:
            warn(
                f"ERROR [menu {self._name}] : text size is too small, min is 1"
            )
            return
        self._text_size = size
        self.set_max_width()

    def set_max_ticks(self, sec: float = 1.) -> None:
        """
        sets maximum ticks of animation depending of the duration of the animation

        Parameters
        ----------
            sec : float, (optional)
                duration of the animation
                defaults to 1
        """
        fps = self._renderer.fps
        self.max_ticks = round(fps * sec)

    def set_max_width(self, cap: int = None) -> None:
        """
        sets maximum width of the menu depending on its text\\
        will not exceed ``cap`` but will be greater than 100

        Parameters
        ----------
            cap : int
                max width
        """
        if cap is None:
            cap = self._renderer.win_width
        if cap <= 100:
            self.width = 100
            return
        max_width = 100
        width = 0

        self._renderer.push()
        self._renderer.text_size = self.text_size
        font = self._renderer.font
        for item in self._all_items:
            label = font.render(item, True, (0, 0, 0))
            width = label.get_width()
            if width > max_width:
                if width > cap:
                    max_width = cap
                    break
                max_width = width
        label = font.render(self.name, True, (0, 0, 0))
        self._namew = label.get_width() + 10
        self._renderer.pop()

        self.width = max_width

    def fold(self) -> None:
        """
        folds the menu : its content becomes hidden\\
        opposite method is ``unfold``
        """
        if self.is_fold or self.is_playing:
            warn(
                f"WARNING [menu {self._name}] : menu is already fold or is currently animated, nothing changed"
            )
            return
        self._is_fold = True
        self._is_playing = True
        self.set_max_ticks()

    def unfold(self) -> None:
        """
        unfolds the menu : its content is revealed\\
        opposite method is ``fold``
        """
        if (not self.is_fold) or self.is_playing:
            warn(
                f"WARNING [menu {self._name}] : menu is already unfold or is currently animated, nothing changed"
            )
            return
        self._is_fold = False
        self._is_playing = True
        self.set_max_ticks()

    def new_items(self, **kwargs) -> None:
        """
        adds items inside the menu

        Keywords Arguments
        ------------------
            * : str
                name of the buttons on the menu, in order
                must be linked to a python function
        """
        for k, v in kwargs.items():
            if k in self._all_items:
                warn(
                    f"ERROR [menu {self._name}] : {k} is already an item, this item was not added, try update_items for this specific item"
                )
                continue
            self._all_items.append(k)
            self._all_actions.append(v)
        self.set_max_width()

    def update_items(self, **kwargs) -> None:
        """
        modifies items inside the menu

        Keywords Arguments
        ------------------
            * : str
                name of the buttons on the menu, in order
                must be linked to a python function
        """
        for k, v in kwargs.items():
            if k not in self._all_items:
                warn(
                    f"ERROR [menu {self._name}] : {k} is not an existing item so it was not modified, try new_items for this specific item"
                )
                continue
            i = self._all_actions.index(k)
            self._all_actions[i] = v
        self.set_max_width()

    def trigger(self, index: int):
        """
        triggers action of designated index\\
        proper index can be found using ``collide``

        Parameters
        ----------
            index : int
                index of item
        """
        if index >= len(self._all_items):
            warn(
                f"ERROR [menu {self._name}] : index {index} does not correspond to a valid item"
            )
            return
        return self._all_actions[index]()

    def update_state(self, pos: tuple) -> None:
        """
        updates the state of the engine based on the position of the mouse

        Parameters
        ----------
            pos : tuple
                mouse position
        """
        renderer = self._renderer

        if self.is_fold:
            x, y = (5, renderer.win_width - 20)[self.side == RIGHT], 5
            if x <= pos[0] <= x + 15 and y <= pos[1] <= y + 15:
                self.unfold()

        x, y = (5 + self.width,
                renderer.win_width - 20 - self.width)[self.side == RIGHT], 5
        if x <= pos[0] <= x + 15 and y <= pos[1] <= y + 15:
            self.fold()

    def draw(self) -> None:
        """
        draws the menu on the screen\\
        different rendering depending on its folding state
        """
        renderer = self._renderer

        if self.is_fold:
            x, y = -1, 5
            x0, xw = -1, -1
            if self.side == RIGHT:
                k = renderer.win_width - 20
                x = _map(self.tick_count, 0, self.max_ticks, k - self.width, k)
                x0 = x
                xw = x0 + 30
            if self.side == LEFT:
                x = _map(self.tick_count, 0, self.max_ticks, self.width, 5)
                x0 = x - self.width
                xw = x - self._namew

            renderer.push()

            if self._is_playing:
                if self.has_background:
                    if self.background is None:
                        renderer.fill = renderer.win_bg
                    else:
                        renderer.fill = self.background
                    # renderer.no_stroke()
                    l = len(self._all_items)
                    h = 15 + l*30
                    if self.length is not None:
                        h = self._length
                    pygame.draw.rect(renderer._window, renderer.fill,
                                     (x0, y, self.width, h))
                    # renderer.rect((x0, y), self.width, h)

                # renderer.no_fill()
                # renderer.stroke = self.color
                # renderer.stroke_weight = 1
                renderer.text_size = 15
                renderer.text_color = self.text_color

                renderer.text(xw, y, self.name)
                for i, item in enumerate(self._all_items):
                    renderer.text(x0, y + 30 + (i*30), item)
                    pygame.draw.line(renderer._window, self.color,
                                     (x0, y + 45 + (i*30)),
                                     (x0 + self.width, y + 45 + (i*30)), 1)
                    # renderer.line((x0, y + 45 + (i*30)), (x0 + self.width, y + 45 + (i*30)))

            # renderer.no_fill()
            # renderer.stroke = self.color
            # renderer.stroke_weight = 2
            for _ in range(3):
                pygame.draw.line(renderer._window, self.color, (x, y),
                                 (x + 15, y), 2)
                # renderer.line((x, y), (x + 15, y))
                y += 5
            renderer.pop()

            self.animate()
            return

        x, y = -1, 5
        x0, xw = -1, -1
        if self.side == RIGHT:
            k = renderer.win_width - 20 - self.width
            x = _map(self.tick_count, 0, self.max_ticks,
                     renderer.win_width - 20, k)
            x0 = x
            xw = x0 + 30
        elif self.side == LEFT:
            k = 5 + self.width
            x = _map(self.tick_count, 0, self.max_ticks, 5, k)
            x0 = x - self.width
            xw = x - self._namew

        renderer.push()
        renderer.text_size = self.text_size
        if self.has_background:
            if self.background is None:
                renderer.fill = renderer.win_bg
            else:
                renderer.fill = self.background
            # renderer.no_stroke()
            l = len(self._all_items)
            h = 15 + l*30
            if self.length is not None:
                h = self._length
            pygame.draw.rect(renderer._window, renderer.fill,
                             (x0, y, self.width, h))
            # renderer.rect((x0, y), self.width, h)

        # renderer.no_fill()
        # renderer.stroke = self.color
        # renderer.stroke_weight = 2
        renderer.text_size = 15
        renderer.text_color = self.text_color

        pygame.draw.line(renderer._window, self.color, (x, y),
                         (x + 15, y + 15), 2)
        pygame.draw.line(renderer._window, self.color, (x, y + 15),
                         (x + 15, y), 2)
        # renderer.line((x, y), (x + 15, y + 15))
        # renderer.line((x, y + 15), (x + 15, y))

        # renderer.stroke_weight = 1

        renderer.text(xw, y, self.name)
        for i, item in enumerate(self._all_items):
            renderer.text(x0, y + 30 + (i*30), item)
            pygame.draw.line(renderer._window, self.color,
                             (x0, y + 45 + (i*30)),
                             (x0 + self.width, y + 45 + (i*30)), 1)
            # renderer.line((x0, y + 45 + (i*30)), (x0 + self.width, y + 45 + (i*30)))
        renderer.pop()

        self.animate()

    def animate(self) -> None:
        """
        go trough animation when unfolding or folding\\
        does one frame of animation based on tick_count\\
        ends animation if needed
        """
        self.tick_count += self.is_playing
        if self.tick_count >= self.max_ticks - 1:
            self.max_ticks = 0
            self.tick_count = 1
            self.is_playing = False

    def collide(self, pos: tuple[int, int]) -> int:
        """
        gets index of item under pos\\
        might be None
        """
        if self.is_fold:
            return

        x, y = pos
        x0, y0 = -1, 5
        if self.side == RIGHT:
            x0 = self._renderer.win_width - 20 - self.width
        elif self.side == LEFT:
            x0 = self.width

        l = len(self._all_items)
        if x0 <= x <= x0 + self.width and y0 <= y <= y0 + 15 + l*30:
            y -= y0 + 15
            y //= 30
            return y



import os

__all__ = ["warn", "error_console_set_soft", "error_console_load_soft"]


def move(y, x):
    """
    moves the cursor to the ``(y, x)`` location in the terminal
    """
    print("\033[%d;%dH" % (y, x))


def cls():
    """
    clears terminal, os independant
    """
    os.system("cls" if os.name == "nt" else "clear")


class ErrorHandler:
    """
    ErrorHandler
    ============
    An invisible class to handle warnings. Phoenyx uses it for warnings and soft errors.
    """
    def __init__(self) -> None:
        """
        new ErrorHangler instance
        """
        cls()
        print("Phoenyx -- errors console")
        move(2, 0)
        self.all_errors: dict[str, int] = dict()
        self.is_soft = True

    def warn(self, msg: str) -> None:
        """
        adds ``msg`` to the error log
        """
        try:
            self.all_errors[msg] += 1
        except KeyError:
            self.all_errors[msg] = 1

        if self.is_soft:
            self.display_all()
        else:
            print(msg, end="\n")

    def display_all(self) -> None:
        """
        display all errors
        """
        i = 1
        for k, v in self.all_errors.items():
            i += 1
            move(i, 0)
            print(f"\r{k} ({v})", end="", flush=True)


err: ErrorHandler  # default for Phoenyx


def error_console_load_soft() -> None:
    """
    initializes Error Handeler\\
    very usefull for debuging
    """
    global err
    err = ErrorHandler()


def error_console_set_soft(flush: bool) -> None:
    """
    sets wether or not the errors are displayed the "pretty" way

    Parameters
    ----------
        flush : bool
            True means errors will not turn into spam
    """
    global err
    try:
        err.is_soft = flush
    except NameError:
        print(
            f"ERROR [error handler] : error handler is not running, you might consider adding ``error_handler_load_soft()`` in the previous line"
        )


def warn(msg: str) -> None:
    """
    warns the user\\
    does not repeat previous warnings if error handler has been initialized

    Parameters
    ----------
        msg : str
            the string that represent the message
    """
    global err
    try:
        err.warn(msg)
    except NameError:
        print(msg, end="\n")
